#include <EtherCard.h>
#include <OneWire.h>
#include <DallasTemperature.h>
#include <EEPROM.h>
#include <Ultrasonic.h>
#include <MemoryFree.h>
#include"Default.h"
#include <avr/pgmspace.h>

//#define DEBUG 1                       // incl. seriellem Monitor zum debuggen
#define SONIC 1                       // incl. Code für Ultraschall-Sensor
#define HTTPINTERVAL 5000             // starting interval for HTTP posts and increment if nothing has to be done 
#define COMPINTERVAL 1000             // compute interval for the PID
/* **********************************************************************************************
    
     ######                           #####
     #     #  #####   ######  #    # #     #   ####   #    #   #####  #####    ####    #
     #     #  #    #  #       #    # #        #    #  ##   #     #    #    #  #    #   #
     ######   #    #  #####   #    # #        #    #  # #  #     #    #    #  #    #   #
     #     #  #####   #       # ## # #        #    #  #  # #     #    #####   #    #   #
     #     #  #   #   #       ##  ## #     #  #    #  #   ##     #    #   #   #    #   #
     ######   #    #  ######  #    #  #####    ####   #    #     #    #    #   ####    ######

                                 #    #    ##    #    #   ####
                                 ##   #   #  #   ##   #  #    #
                                 # #  #  #    #  # #  #  #    #
                                 #  # #  ######  #  # #  #    #
                                 #   ##  #    #  #   ##  #    #
                                 #    #  #    #  #    #   ####

   *
   * Author    : H.Herrmannsdörfer
   * Created   : 04/2014
   *
   * Changes   : 15.5.2014 Added 3 PID controllers to control the 3 relay outputs
   *                       see https://learn.adafruit.com/sous-vide-powered-by-arduino-the-sous-viduino/the-whole-enchilada
   *
   *             16.11.2014 Complete rewrite to use the PID controller and first tests 
   *
   ****************************************************************************/

// global variables
byte            relaisstate = 0;                        // akt. Status der Relais-Ausgaenge (bitkodiert , RELAIS1 = 0, RELAIS2=1, ...)
float           akttempsC[4] = { -50.0, -50.0, -50.0, -50.0 };
float           settemp;                                // Aktuelle Zieltemperatur
float           firstoff;                               // Ab welcher Temperaturdifferenz wird vorher abgeschaltet
unsigned long   laststeptime;                           // Anfangszeit der Antast-Steps
unsigned long   cycletime;                              // Anfangszeit der Antast-Steps
unsigned long   heattime;                               // Anfangszeit der Antast-Steps
byte            aktorder;                               // aktueller Auftrag vom Server 
boolean         sendconfirm = false;                    // muss eine Bestätigung an den Server gesendet werden ?
boolean         stepmode = false;                       // wurde die Zieltemperatur schon erreicht?
char            sendmessage[180] = "";                  // setzen, um Meldungen an den Server zu schicken
static long     lastConnectionTime;                     // last time you connected to the server, in milliseconds
static long     lastComputeTime;                        // last time you connected to the server, in milliseconds
unsigned long   postingInterval = HTTPINTERVAL;         // delay between updates, in milliseconds
byte            raiseIntervalCount = 0;                 // counter to slow down delay raising 

/*-----( Declare objects )-----*/
OneWire oneWire(SENSORPIN);                             // Setup a oneWire instance to communicate with any OneWire devices
DallasTemperature sensors(&oneWire);                    // Pass our oneWire reference to Dallas Temperature. 

#ifdef SONIC
Ultrasonic ultrasonic(SONIC_TRG, SONIC_ECHO);           // Objekt für Ultraschall-Sensor
#endif

// Local settings
SettingsData  settings  = {
    1,
    { 192, 168, 100, 150 },
    { 192, 168, 100, 1   },
    1,
    0,
    { 
      { 0x28, 0x04, 0xD5, 0x4D, 0x04, 0x00, 0x00, 0xAD }, 
      { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
      { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
      { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
    },
    {   0,   0,   0,   0 },
    850.0,
    0.5,
    0.1,
    2.0,
    0.5,
    30, 
    300
  };

// ethernet mac address - must be unique on your network
static byte mymac[] = { 0x74,0x69,0x69,0x2d,0x30,0x31 };
byte Ethernet::buffer[400];
char webpath[] PROGMEM = "/Brauerei/ard_conn.php?";
char website[] PROGMEM = "rover";

// Ausgabewert für die Regelung der Heizrelais
volatile long onTime = 0;            // PID Ergebnis

// 10 second Time Proportional Output window
const int WINDOWSIZE = 5000; 
unsigned long windowStartTime;




// ================================================
// SETUP
// ================================================
void setup(void) {
  uint8_t epromtest;

#ifdef DEBUG
  Serial.begin(9600);
  Serial.println("BrewFlobber ... setup");
#endif  

  // have the settings ever been written ? read the first byte of the settings to check 
  // that. otherwise take the default values
  epromtest = EEPROM.read(0);
  if (epromtest != 0 && epromtest != 255) {
    EEPROM_readAnything(0, settings);
#ifdef DEBUG
    Serial.println("Read settings from EEPROM...");
#endif
  } else {
    resetValues();
  }  

#ifdef DEBUG
  if (ether.begin(sizeof Ethernet::buffer, mymac, ETHERPIN) == 0) 
    Serial.println( "Failed to access Ethernet controller");
#else
  ether.begin(sizeof Ethernet::buffer, mymac, ETHERPIN);
#endif
//  if (!ether.dhcpSetup()) {
//    Serial.println("DHCP failed, using static configuration");
#ifdef DEBUG
    if (!ether.staticSetup(settings.src_ip, settings.dst_ip))
      Serial.println("Failed to set IP address");
#else
    ether.staticSetup(settings.src_ip, settings.dst_ip);
#endif
//  }

  ether.copyIp(ether.dnsip, settings.dst_ip);

  if (!ether.dnsLookup(website)) {
#ifdef DEBUG
    Serial.println("DNS failed, setting fixed srvIP ");
#endif
    ether.copyIp(ether.hisip, settings.dst_ip);
  }
  
  // print actual settings
/*
#ifdef DEBUG
  Serial.println("Setup SOLL");
  Serial.println("-------------------------------------------");
  Serial.print(F("Anlage :"));
  Serial.println(settings.anl_no);
  Serial.print(F("Heizer :"));
  Serial.println(settings.heatercount);
  Serial.print(F("Wasser :"));
  Serial.println(settings.max_water);
  ether.printIp("SrcIP  :", settings.src_ip);   
  ether.printIp("DstIP s :", settings.dst_ip);
  ether.printIp("DstIP i :", ether.hisip);
  Serial.println("-------------------------------------------");

  Serial.println(F("Setup IST"));
  Serial.println(F("-------------------------------------------"));
  ether.printIp("IP       : ", ether.myip);
  ether.printIp("DNS      : ", ether.dnsip);  
  ether.printIp("SRV ist  : ", ether.hisip);
  Serial.println(F("-------------------------------------------"));
#endif
*/

#ifdef DEBUG
  Serial.println(F("Settings"));
  Serial.print(F("Anlage     :")); Serial.println(settings.anl_no);
  Serial.print(F("Heizungen  :")); Serial.println(settings.heatercount);
  Serial.print(F("Max. Wasser:")); Serial.println(settings.max_water);
  
  // Sensor-Abweichungen
  Serial.print(F("Abweichung0:")); Serial.println(settings.deviation[0]);
  
  Serial.print(F("FirstOff   :")); Serial.println(settings.FirstOff);
  Serial.print(F("SecondOn   :")); Serial.println(settings.SecondOn);
  Serial.print(F("HeizZeit   :")); Serial.println(settings.HeizZeit);
  Serial.print(F("ZyklusZeit :")); Serial.println(settings.ZyklusZeit);
#endif

  if (settings.Sensors[0][0] != 0) sensors.setResolution(settings.Sensors[0], 9); 
  if (settings.Sensors[1][0] != 0) sensors.setResolution(settings.Sensors[1], 9); 
  if (settings.Sensors[2][0] != 0) sensors.setResolution(settings.Sensors[2], 9); 
  if (settings.Sensors[3][0] != 0) sensors.setResolution(settings.Sensors[3], 9); 
  
  // Relais-Pins einstellen
  pinMode(RELAYPIN_1, OUTPUT);
  pinMode(RELAYPIN_2, OUTPUT);
  pinMode(RELAYPIN_3, OUTPUT);
  pinMode(MIXERPIN, OUTPUT);
  digitalWrite(RELAYPIN_1, RELAY_OFF);
  digitalWrite(RELAYPIN_2, RELAY_OFF);
  digitalWrite(RELAYPIN_3, RELAY_OFF);
  digitalWrite(MIXERPIN  , RELAY_OFF);

  calcHysteresis();

#ifdef DEBUG
  Serial.println(F("Setup"));
  Serial.print(F("HeizZeit   :")); Serial.println(heattime);
  Serial.print(F("ZyklusZeit :")); Serial.println(cycletime);
#endif

  lastConnectionTime = millis();
  lastComputeTime = millis();
}



void loop(void) {
  long akttime;

  ether.packetLoop(ether.packetReceive());
  akttime = millis();
  
  /* ===========================================================
  Bekannte Temperaturen auslesen
  =========================================================== */
  getTemperatures(); 

  // Wenn eine Temperatur geregelt werden soll, dann die Heizungen anhand des PID einstellen
  if (aktorder == 1 && allowHeater()) {
    // Nicht zu oft messen
    if (akttime > lastComputeTime) {
      onTime = 0;                                                         // erstmal nicht heizen
      
#ifdef DEBUG
        Serial.print(F("StepMode :"));        Serial.println(stepmode);
        Serial.print(F("Laststeptime :"));    Serial.println(laststeptime);
        Serial.print(F("Akttime      :"));    Serial.println(akttime);
        Serial.print(F("Cycletime    :"));    Serial.println(cycletime);
        Serial.print(F("Heattime     :"));    Serial.println(heattime);
#endif
      if (!stepmode) {
        if (akttempsC[0] + firstoff < settemp) {
          onTime = WINDOWSIZE;                                            // Temperatur wurde noch nie erreicht => volle Kanüle heizen
        } else {
          stepmode = true;                                                // Übergang in den Takt-Modus
          laststeptime = akttime + cycletime;
        }
      } else {
        if (akttime < laststeptime) {                                   // nächsten Zyklus beginnen ? 
          // Ist die Temperatur noch zu niedrig ?
          if ((akttempsC[0] + settings.SecondOn <= settemp) &&
              (akttime > laststeptime - heattime)) {                    // Heizzeit am Ende innerhalb des Zyklus?
#ifdef DEBUG
        Serial.println(F("Heizung an"));
#endif
            onTime = WINDOWSIZE; 
          }
        } else {
          laststeptime = akttime + cycletime;
        }
      }
      
      // Takt-Modus ausschalten, wenn die Temperatur zu stark gefallen ist
      if (akttempsC[0] + firstoff + settings.SecondOn < settemp) {
#ifdef DEBUG
        Serial.print(F("AkttempsC :"));    Serial.println(akttempsC[0]);
        Serial.print(F("FirstOff  :"));    Serial.println(firstoff);
        Serial.print(F("SecondOn  :"));    Serial.println(settings.SecondOn);
        Serial.print(F("Sum       :"));    Serial.println(akttempsC[0] + firstoff + settings.SecondOn);
        Serial.print(F("SetTemp   :"));    Serial.println(settemp);
        Serial.println(F("StepMode AUS"));
#endif
        onTime = WINDOWSIZE; 
        stepmode = false;
      }

      lastComputeTime = akttime + COMPINTERVAL;
    }
    // Anhand des Ergbnisses das/die Relais steuern und berechnen, ob das Rührwerk laufen soll
    if (onTime > 0) relaisstate |= 8;        // set mixer-bit
    if (relaisstate && 8) digitalWrite(MIXERPIN, RELAY_ON);
  } else {
    onTime = 0;
    relaisstate = 0;
    digitalWrite(MIXERPIN, RELAY_OFF);
  }

  DriveOutput();
  
  if(akttime > lastConnectionTime) {
    // wenn sowieso nichts zu tun ist, dann langsam die Request-Intervalle erhöhen
    if (aktorder != 1) {
      if (postingInterval < 12000000) {
        if (raiseIntervalCount < 5) {
          raiseIntervalCount++;
        } else {
          postingInterval += HTTPINTERVAL;     // request-Intervalle erhöhen
          raiseIntervalCount = 0;
        }
      }
    } else {
      postingInterval = HTTPINTERVAL;
    }
    
    lastConnectionTime = akttime + postingInterval;
    httpRequest();
  } 

  delay(50);
}





/* ===========================================================
   allowHeater
   
   Wenn ein Mindestabstand gesetzt ist, dann muss er auch erreucht sein, bevor
   man heizen darf. Ansonsten darf man immer heizen.
   =========================================================== */
boolean allowHeater() {
  // wenn kein Sensor vorhanden ist, dann sollte man auch nicht heizen
  if (akttempsC[0] < 0.0) {
    return false;
#ifdef SONIC    
  // wenn ggf. kein Wasser drin ist
  } else if ((settings.max_water > 0) && (ultrasonic.Ranging(CM) > settings.max_water)) {
    return false;
#endif
  } else {
    return true;
  }
}



/* ===========================================================
   getTemperatures
   
   Temperaturen von allen bekannten Sensoren abfragen.
   =========================================================== */
void getTemperatures() {
  sensors.requestTemperatures();
  if (settings.Sensors[0][0] != 0) akttempsC[0] = sensors.getTempC(settings.Sensors[0]);
  if (settings.Sensors[1][0] != 0) akttempsC[1] = sensors.getTempC(settings.Sensors[1]); 
  if (settings.Sensors[2][0] != 0) akttempsC[2] = sensors.getTempC(settings.Sensors[2]); 
  if (settings.Sensors[3][0] != 0) akttempsC[3] = sensors.getTempC(settings.Sensors[3]); 
}



/* ===========================================================
   Callback-Funktion für browseUrl. Hier landet die Antwort.
   =========================================================== */
static void response_callback(byte status, word off, word len) { 
  const char* cbuf = ((const char*) Ethernet::buffer + off);
  float newsettemp;

  char *body = strstr(cbuf, "\r\n\r\n"); // search for CR NL CR NL 
  if (body) {                           // found 
    body += 4;                          // skip CR NL CR NL 
  } else {                              // not found 
    body = strstr(cbuf, "\n\n");        // search for NL NL 
    if (body) {                         // found 
      body += 2;                        // skip NL NL 
    } 
  }

  if (body) {                           // blank line found, body points to just after it 
    aktorder = body[0];                 // convert string to unsigned long 
    
    // sofort auf minimales Postingintervall umschalten, wenn es was zu tun gibt
    if (aktorder != 0) {
      postingInterval = HTTPINTERVAL;
      lastConnectionTime = millis() + postingInterval;
    }
    
    // anhand des neuen Kommandos verzweigen
    switch(aktorder) {
      case 0 :				// do nothing
        sendconfirm = false;       
        sendmessage[0] = 0;
        settemp = 0.0;                  // aktuell Zieltemperatur löschen
#ifdef DEBUG
        Serial.println(F("--"));
#endif
        stepmode = false;               // Regelung neu initialisieren
        break;
      case 1:				// Bringe Topf auf Temp x
        newsettemp = body[1] * 1.0;     // neue Zieltemperatur auslesen
        if (newsettemp != settemp) {
          stepmode = false;             // Regelung neu initialisieren
        }
        settemp = newsettemp;           // neue Zieltemperatur übernehmen
        calcHysteresis();               // ggf. die Hysterese-Werte anpassen
        windowStartTime = millis();
        postingInterval = HTTPINTERVAL; // sofort auf minimales Postingintervall umschalten 
        break;
//      case 2:				// Schalte Relais direkt
//        Serial.println("Schalte Relais direkt");       break;
      case 16:				// lerne Adresse für Sensor x
#ifdef DEBUG
        if (lookUpSensorNoX(body[1])) {
          Serial.print(F("gefundene Adresse :"));
          Serial.println(sendmessage);
        } else {
          Serial.println(F("nix gefunden :-("));
        };
#else
        lookUpSensorNoX(body[1]);
#endif
        break;
      case 17:				// lerne Messabweichung für Sensor x
        settings.deviation[body[1]] = body[2]; 
        sendconfirm = true;
        break;
      case 18:				              // lerne Anzahl Heizungen
        settings.heatercount  = body[1];
        sendconfirm = true;
        break;        
      case 19:				              // lerne Anlagennummer
        settings.anl_no = body[1];
        sendconfirm = true;
        break;
      case 20:				              // lerne Mindestfüllstand
#ifdef DEBUG
        Serial.println("lerne Mindestfüllstand");
#endif
#ifdef SONIC
        settings.max_water = ultrasonic.Ranging(CM);
#endif
        break;
      case 22:                                        // speichere Einstellungen im EEPROM
        EEPROM_writeAnything(0, settings);
        sendconfirm = true;
        break;
      case 32:
#ifdef DEBUG
        Serial.println("reset");
#endif
        break;
      case 26:                                        // setze Hysterese-Werte
        // $anlagen[$aktanlage]->actorder = chr(0x1A) . chr($_POST['firstoff']*10) . chr($_POST['secondon']*10) . chr($_POST['heizzeit']);
        settings.FirstOff = body[1] / 10.0;
        settings.SecondOn = body[2] / 10.0;
        settings.HeizZeit = (unsigned long) body[3] * 1.0;
        calcHysteresis();
        break;
      default:
        break;
    }                    // switch()
  }                      // if (client.available())
}

/* ===========================================================
   httpRequest
   
   Temperaturen und Relaiszustand der Anlage an den Server senden 
   und neue Befehle abholen.
   =========================================================== */
void httpRequest() {
  char srequest[200] = {0};        // zusammengebastelter Request
  char string_temp[10][8] = {"0.0","0.0","0.0","0.0"};
  int  fmem = 0;
  
  if (sendconfirm) {                 // muss noch eine Bestätigung gesendet werden?
    sprintf(srequest, "a=%d&confirm=1", settings.anl_no);
#ifdef DEBUG
  Serial.print(F("conf "));
#endif
    sendconfirm = false;
  } else if (sendmessage[0] != 0) {
    sprintf(srequest, "a=%d&message=%s", settings.anl_no, sendmessage);
    sendmessage[0] = 0;
  } else {                           // normales Paket. Temperaturen senden
    dtostrf(akttempsC[0], 4, 1, string_temp[0]);
    dtostrf(akttempsC[1], 4, 1, string_temp[1]);
    dtostrf(akttempsC[2], 4, 1, string_temp[2]);
    dtostrf(akttempsC[3], 4, 1, string_temp[3]);
    dtostrf(heattime/1000, 4, 0, string_temp[4]);    
    dtostrf(firstoff, 4, 1, string_temp[5]);    
    dtostrf(settemp, 4, 1, string_temp[6]);    
    fmem = freeMemory();

    // Normalen Request an den Server senden
//    sprintf(srequest, "a=%d&t1=%s&t2=%s&t3=%s&t4=%s&r=%d&set=%s&foff=%s&son=%s&hz=%d&zz=%d&fm=%d", 
    sprintf(srequest, "a=%d&t1=%s&t2=%s&t3=%s&t4=%s&r=%d&set=%s&step=%d&ht=%s&fo=%s&fm=%d", 
            settings.anl_no, &string_temp[0], &string_temp[1], 
            &string_temp[2], &string_temp[3], relaisstate, 
            &string_temp[6], stepmode,
            &string_temp[4], &string_temp[5], 
//          settings.HeizZeit, settings.ZyklusZeit,
            fmem);
  }
#ifdef DEBUG
  Serial.print(F("request :"));
  Serial.println(srequest);
#endif
  //ether.browseUrl(PSTR("/Brauerei/connect.php?"),srequest, website, response_callback);
  ether.browseUrl(webpath,srequest, website, response_callback);
}



// ===========================================================
// lookUpSensorNoX
// 
// Neuen Sensor-Adresse für einen der 4 Sensoren finden. Die 
// Adressen werden mit den bekannten Adressen verglichen, damit
// man den neuen auch findet.
// ===========================================================
boolean lookUpSensorNoX(byte sensorNo) {
  byte foundaddress[8];
  int  i=0;
  byte aktaddress = 0;
  boolean ok = false, all_ok = true;

  while (oneWire.search(foundaddress)) {                              // solange es neue Adressen gibt    
    if (foundaddress[0] == 0x10 || foundaddress[0] == 0x28)  {        // 0x10 = DS18S20 / 0x28 = DS18B20
      if (OneWire::crc8(foundaddress, 7) == foundaddress[7]) {        // CRC der Adresse ist OK
        for (aktaddress=0;aktaddress<3;aktaddress++) {                // mit den bekannten Adressen vergleichen
          ok = false;
          if (aktaddress != sensorNo) {   // Den gesuchten Eintrag nicht mitvergleichen
            for (i=0;i<8;i++) {
              if (settings.Sensors[aktaddress][i] != foundaddress[i]) 
                ok = true;    // there's a difference in the adress, so it doesn't match a known address
                i = 8;
            }
            if (!ok) all_ok = false;
          }
        }
      }
    }               // end if foundaddress[0] == ...
  }                 // end while
  
  // when everything is ok, take the address as new sensor address  
  if (all_ok) {
    for (i=0;i<8;i++) {
      settings.Sensors[sensorNo][i] = foundaddress[i];
    }

    sensors.setResolution(settings.Sensors[sensorNo], 9); // initialize sensor
    
    // Antwort mit der Adresse an den Server senden
    sprintf(sendmessage, "Sensor+%d+Address+%2X:%2X:%2X:%2X:%2X:%2X:%2X:%2X", sensorNo+1,
          foundaddress[0], foundaddress[1], foundaddress[2], foundaddress[3],
          foundaddress[4], foundaddress[5], foundaddress[6], foundaddress[7]);
  }
  return all_ok;
}



// ===========================================================
// calcHysteresis
// 
// Anhand der eingestellten Zieltemperatur die Timingwerte berechnen.
// Ab 70° aufwärts müssen die Schwellwerte angepasst werden.
// =========================================================== 
void calcHysteresis() {
  if (settemp < 70) {
    cycletime = settings.ZyklusZeit * 1000.0;
    heattime  = settings.HeizZeit * 1000.0;
    firstoff  = settings.FirstOff;
  } else if (settemp < 80) {
    cycletime = settings.ZyklusZeit * 1000.0;
    heattime  = settings.HeizZeit * 1500.0;       // 50% länger zwischenheizen
    firstoff  = settings.FirstOff - 0.5;          // 0.5° später abschalten
  } else if (settemp > 80) {
    cycletime = settings.ZyklusZeit * 1000.0;
    heattime  = settings.HeizZeit * 2000.0;       // 100% länger zwischenheizen
    firstoff  = settings.FirstOff - 1.5;          // 1.5° später abschalten
  } else if (settemp > 95) {
    cycletime = settings.ZyklusZeit * 1000.0;
    heattime  = settings.HeizZeit * 2000.0;       // 100% länger zwischenheizen
    firstoff  = settings.FirstOff - 1.5;          // 1.5° später abschalten
  }
}


// ===========================================================
// resetValues
// 
// Alle Einstellungswerte in den Grundzustand zurücksetzen. Wird
// auch beim Setup aufgerufen, wenn nichts im EEPROM gespeichert 
// war.
// =========================================================== 
void resetValues() {
  settings.anl_no        = 1;
  settings.heatercount   = 1;
  settings.max_water     = 0;
  
  // Sensor addressen
  settings.Sensors[0][0] = 0x28;  settings.Sensors[0][1] = 0x04;  settings.Sensors[0][2] = 0xD5;  settings.Sensors[0][3] = 0x4D;
  settings.Sensors[0][4] = 0x04;  settings.Sensors[0][5] = 0x00;  settings.Sensors[0][6] = 0x00;  settings.Sensors[0][7] = 0xAD;

  settings.Sensors[1][0] = 0;  settings.Sensors[2][0] = 0;  settings.Sensors[3][0] = 0;
  
  // Sensor-Abweichungen
  settings.deviation[0]  = 0;  settings.deviation[1]  = 0;  settings.deviation[2]  = 0;  settings.deviation[3]  = 0;
  
  settings.src_ip[0] = 192;  settings.src_ip[1] = 168;  settings.src_ip[2] = 100;  settings.src_ip[3] = 150;
  settings.dst_ip[0] = 192;  settings.dst_ip[1] = 168;  settings.dst_ip[2] = 100;  settings.dst_ip[3] = 1;
  
  // PID-parameter für einen einfachen Einkocher 
  settings.Kp = 600;
  settings.Ki = 0.5;
  settings.Kd = 0.1;

  settings.FirstOff    = 2.0;           // Erste Hysterese: Zuerst ab diesem Wert (unter Zielwert) mal abschalten und die Restwärme nutzen
  settings.SecondOn    = 0.5;           // Wenn die Temp. nach ZyklusZeit unter diesen Schwellwert gesunken ist, dann die Heizung wieder innerhalb der ZyklusZeit für die HeizZeit einschalten.
  settings.HeizZeit    = 30;            // So viele Sekunden einschalten
  settings.ZyklusZeit  = 300;           // innerhalb dieser Zeitspanne
}



// ************************************************
// Called by ISR every 15ms to drive the output
// ************************************************
void DriveOutput() {  
  long now = millis();
  
  // Set the output
  // "on time" is proportional to the PID output
  if ((now - windowStartTime) > WINDOWSIZE) { //time to shift the Relay Window
     windowStartTime += WINDOWSIZE;
  }
  // Das > 500 ist noch von der PID-Regelung übrig, stört aber ohne PID auch nicht
  if((onTime > 500) && (onTime > (now - windowStartTime))) {
     digitalWrite(RELAYPIN_1,RELAY_ON); relaisstate |= B1;
     if (settings.heatercount > 1) {
       digitalWrite(RELAYPIN_2,RELAY_ON);
       relaisstate |= B10;
     }
     if (settings.heatercount > 2) {
       relaisstate |= B100;
       digitalWrite(RELAYPIN_3,RELAY_ON);
     }
  } else {
     digitalWrite(RELAYPIN_1,RELAY_OFF);
     relaisstate &= B1110;
     if (settings.heatercount > 1) {
       relaisstate &= B1101;
       digitalWrite(RELAYPIN_2,RELAY_OFF);
     }
     if (settings.heatercount > 2) {
       relaisstate &= B1011;
       digitalWrite(RELAYPIN_3,RELAY_OFF);
     }
  }
}

