#ifndef DEFAULT_H
#define DEFAULT_H

#if ARDUINO >= 100
#include "Arduino.h"       // for delayMicroseconds, digitalPinToBitMask, etc
#else
#include "WProgram.h"      // for delayMicroseconds
#include "pins_arduino.h"  // for digitalPinToBitMask, etc
#endif

#define RELAY_ON    0x1                // turn RELAY ON
#define RELAY_OFF   0x0                // turn RELAY OFF

typedef struct {
  byte           anl_no;               // Anlagennummer
  byte           src_ip[4];            // eigene IP adresse
  byte           dst_ip[4];            // IP address des Servers
  byte           heatercount;          // Anzahl Heizungen
  long           max_water;            // Maxwert des Ultraschall-Sensors, ab dem der Kessel nicht 
                                       // mehr als leer betrachtet wird. (Messung vom Deckel aus ! )
                                       // Alles kleinere ist also ein vollerer Kessel.
  DeviceAddress  Sensors[4];           // OneWire-Adressen der 4 Temp.-Sensoren
  byte           deviation[4];         // Messabweichungen der Sensoren
  
  // values for the PID
  double         Kp;
  double         Ki;
  double         Kd;
  
  // values for hysteresis mode
  float          FirstOff;             // Erste Hysterese: Zuerst ab diesem Wert (unter Zielwert) mal abschalten und die Restwärme nutzen
  float          SecondOn;             // Wenn die Temp. nach ZyklusZeit unter diesen Schwellwert gesunken ist, dann die Heizung wieder innerhalb der ZyklusZeit für die HeizZeit einschalten.
  int            HeizZeit;             // So viele Sekunden einschalten
  int            ZyklusZeit;           // innerhalb dieser Zeitspanne
} SettingsData;

//#######################
// Pin Definitions
//#######################
#define BUZZERPIN   2           // Pin für den zusätzlichen Buzzerpin (falls angeschlossen)
#define SENSORPIN   3           // Pin für den OneWire-Bus mit den Temp.-Sensoren
#define SONIC_TRG   4           // Pin für Ultraschall-Sensor Trigger
#define SONIC_ECHO  5           // Pin für Ultraschall-Sensor Echo
#define RELAYPIN_1  6           // Pin für Heizungs-SSR 1
#define RELAYPIN_2  7           // Pin für Heizungs-SSR 2
#define RELAYPIN_3  8           // Pin für Heizungs-SSR 3
#define MIXERPIN    9           // Pin für Rührwerks-Relais
#define ETHERPIN   10           // Pin für den Ethernet-Anschluss
//#######################

#include <EEPROM.h>

template <class T> int EEPROM_writeAnything(int ee, const T& value) {
  const byte* p = (const byte*)(const void*)&value;
  unsigned int i;
  for (i = 0; i < sizeof(value); i++)
    EEPROM.write(ee++, *p++);
  return i;
}

template <class T> int EEPROM_readAnything(int ee, T& value) {
  byte* p = (byte*)(void*)&value;
  unsigned int i;
  for (i = 0; i < sizeof(value); i++)
    *p++ = EEPROM.read(ee++);
  return i;
}
#endif
