#ifndef DEFAULT_H
#define DEFAULT_H

#if ARDUINO >= 100
#include "Arduino.h"       // for delayMicroseconds, digitalPinToBitMask, etc
#else
#include "WProgram.h"      // for delayMicroseconds
#include "pins_arduino.h"  // for digitalPinToBitMask, etc
#endif

#define RELAY_ON    0x1                // turn RELAY ON
#define RELAY_OFF   0x0                // turn RELAY OFF

typedef struct {
  byte           anl_no;               // own device identifier
  byte           src_ip[4];            // own IP address of the sender ... 
  byte           dst_ip[4];            // IP address of the server
  byte           heatercount;          // number of heaters if more than one, use a PID algorithm
  long           max_water;            // Maxwert des Ultraschall-Sensors, ab dem der Kessel nicht 
                                       // mehr als leer betrachtet wird. (Messung vom Deckel aus ! )
                                       // Alles kleinere ist also ein vollerer Kessel.
  DeviceAddress  Sensors[4];           // onewire addresses of the 4 possible temperature sensors
  byte           deviation[4];         // measurement deviation per sensor
  
  // values for the possible PIDs
//  double         Kp[3];
//  double         Ki[3];
//  double         Kd[3];
  double         Kp;
  double         Ki;
  double         Kd;
} SettingsData;

//#######################
// Pin Definitions
//#######################
#define BUZZERPIN   2           // Pin für den zusätzlichen Buzzerpin (falls angeschlossen)
#define SENSORPIN   3           // Pin für den OneWire-Bus mit den Temp.-Sensoren
#define SONIC_TRG   4           // Pin für Ultraschall-Sensor Trigger
#define SONIC_ECHO  5           // Pin für Ultraschall-Sensor Echo
#define RELAYPIN_1  6           // Pin für Heizungs-SSR 1
#define RELAYPIN_2  7           // Pin für Heizungs-SSR 2
#define RELAYPIN_3  8           // Pin für Heizungs-SSR 3
#define MIXERPIN    9           // Pin für Rührwerks-Relais
#define ETHERPIN   10           // Pin für den Ethernet-Anschluss
//#######################

#include <EEPROM.h>

template <class T> int EEPROM_writeAnything(int ee, const T& value) {
  const byte* p = (const byte*)(const void*)&value;
  unsigned int i;
  for (i = 0; i < sizeof(value); i++)
    EEPROM.write(ee++, *p++);
  return i;
}

template <class T> int EEPROM_readAnything(int ee, T& value) {
  byte* p = (byte*)(void*)&value;
  unsigned int i;
  for (i = 0; i < sizeof(value); i++)
    *p++ = EEPROM.read(ee++);
  return i;
}
#endif
